package codeline;

import processor.ControlFlow;
import processor.Memory;
import processor.StackFrame;
import processor.Memory.ReadOnlyMemory;
import processor.Memory.Visibility;
import processor.StackFrame.FrameType;

/**
 * the initial line of a for loop of the kind "FOR x := min TO max DO"
 * @author kolja
 *
 */
public abstract class ForLoop extends CodeLine {

    private String loopVar;
	
	public ForLoop( String varName )
	{
		this.loopVar = varName;
	}
	
	@Override
	public ControlFlow runForward(Memory m) {
		boolean declared = false; // prove if it is the first step in the loop
		if( !m.isSomewhereDefined( "line_" + lineId + "_index", Visibility.LOCAL ) )
		{ // first loop step
			m.declare( "line_" + lineId + "_index", minimum( m.createReadOnlyMemory() ), Visibility.LOCAL );
			declared = true;
		}
		if( m.<Integer>read( "line_" + lineId + "_index", Visibility.LOCAL ) > maximum( m.createReadOnlyMemory() ) ) // prove if the loop has finished
		{
			actions.push( (Memory mem) -> {});
			return new ControlFlow( ControlFlow.STEP_OVER ); // don't execute the loop body
		}
		if( declared )
		{
			m.addFrame( new StackFrame( FrameType.LOOP ) );
            m.declare( loopVar, minimum( m.createReadOnlyMemory() ), Visibility.LOCAL ); // set loop variable
			actions.push( (Memory mem) -> {
				mem.removeFrame();
				mem.undeclare( "line_" + lineId + "_index", Visibility.LOCAL );
			} );
		}
		else
		{
			int oldIndex = m.<Integer>read( "line_" + lineId + "_index", Visibility.LOCAL );
			m.write( "line_" + lineId + "_index", oldIndex + 1, Visibility.LOCAL ); // count index down
			if( m.<Integer>read( "line_" + lineId + "_index", Visibility.LOCAL ) > maximum( m.createReadOnlyMemory() ) ) // prove if loop was finished
			{
				StackFrame sf = m.removeFrame(); // remove loop stack
				actions.add( (Memory mem) -> {
	                //m.declare( loopVar, old, false );
					mem.write( "line_" + lineId + "_index", oldIndex, Visibility.LOCAL );
					mem.addFrame( sf ); // restore last loop stack
				});
				return new ControlFlow( ControlFlow.STEP_OVER ); // step out of the loop
			}
			StackFrame old = m.removeFrame(); // fresh stack frame for loop body
			m.addFrame( new StackFrame( FrameType.LOOP ) );
			m.declare( loopVar, (int)m.read( "line_" + lineId + "_index", Visibility.LOCAL ), Visibility.LOCAL ); // update loop variable
			actions.push( (Memory mem) -> {
				mem.removeFrame();
				mem.addFrame( old );
				mem.write( "line_" + lineId + "_index", oldIndex, Visibility.LOCAL );
			});
		}
		return new ControlFlow( ControlFlow.STEP_INTO );
	}
	
	/**
	 * returns the first value of the loop
     * @param m the memory in which the variables are stored
     * @return the starting value
	 */
	protected abstract int minimum( ReadOnlyMemory m );

    
    /**
     * returns the last value of the loop, inclusively
     * @param m the memory in which the variables are stored
     * @return the last value
     */
	protected abstract int maximum( ReadOnlyMemory m );
}
